using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Text;			// for Encoding I/O to a Text file
using System.Threading;		// for Thread

namespace MyNamespace
{
	/// <summary>
	///		Common routines
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class Common
	{
		public Common()		// constructor logic
		{
			nL = Environment.NewLine;
		}

		public static void SingleSpace() // guarantee a single line space in outS
		{
			SingleSpace(MultiID.outS);
		}

		public static void SingleSpace(int outB) // guarantee a single line space in output box
		{
			//MultiID.WriteLn(outB, null);

			//int iL = outB.Text.Length;
			//if (iL > 0)
			//{
			//    if (outB.Text.Substring(iL - 1) != "\n") MultiID.WriteLn(outB, null);
			//}
		}

		public static void DoubleSpace(int outB) // guarantee a double line space in output box
		{
			MultiID.WriteLn(outB, null);

			//int iL = outB.Text.Length;
			//if (iL > 0)
			//{
			//    if (outB.Text.Substring(iL - 1) != "\n") MultiID.WriteLn(outB, null);
			//    if (outB.Text.Substring(outB.Text.Length - 2) != "\n\n") MultiID.WriteLn(outB, null);
			//}
		}

		public static void DoubleSpace() // guarantee a double line space in outS
		{
			DoubleSpace(MultiID.outS);
		}

		public static bool OpenRead(out StreamReader inF, string fName, out string dir)
		// open file for Reading from CurrentDirectory or homeDir
		{
			string stEx1;
			string stEx2 = "";
			string savDir = Environment.CurrentDirectory.ToString();
			if (!quiet) DoubleSpace();
			dir = savDir;
			inF = null;
			if (OpenReadFile(out inF, fName, out stEx1))
			{
				if (!quiet)
				{
					MultiID.WriteDLn("File \"" + fName + "\" opened for reading");
					WCurDir(fName);
				}
				return true;
			}
			if (savDir != homeDir)
			{
				dir = homeDir;
				Directory.SetCurrentDirectory(homeDir);
				if (OpenReadFile(out inF, fName, out stEx2))
				{
					if (!quiet)
					{
						MultiID.WriteLn("File \"" + fName + "\" opened for reading");
						WCurDir(fName);
					}
					Directory.SetCurrentDirectory(savDir);
					return true;
				}
				Directory.SetCurrentDirectory(savDir);
			}
			dir = "";
			MultiID.WriteLn("Cannot open file \"" + fName + "\" for input");
			MultiID.WriteTLn("Not in current directory " + savDir);
			ReportError("OpenRead", stEx1);		// file not found most likely
			if (savDir != homeDir)
			{
				MultiID.WriteTLn("Not in home directory " + homeDir);
				ReportError("OpenRead", stEx2); // file not found most likely
			}
			return false;
		} // OpenRead

		public static bool OpenReadFile(out StreamReader inF, string fName, out String stEx)
		// open file for Reading from CurrentDirectory
		{
			inF = null;
			stEx = "";
			try
			{
				inF = new StreamReader(fName); // open fName for input as inF
				return true;
			}
			catch (Exception ex)
			{
				stEx = ex.Message.ToString();
				return false;
			}
		}

		public static bool OpenWrite(out StreamWriter outF, string fName) // open file for Writing
		{
			outF = null; // outF must be assigned to before control leaves the current method
			try
			{
				// open fName for output as outF
				outF = new StreamWriter(fName, false, Encoding.ASCII); // Appending = false
				if (!quiet)
				{
					MultiID.WriteDLn("File \"" + fName + "\" opened for writing");
					WCurDir(fName);
				}
				return true;
			}
			catch (Exception ex)
			{
				MultiID.WriteDLn("Cannot open file \"" + fName + "\" for output");
				WCurDir(fName);
				ReportError("OpenWrite", ex.Message);
				return false;
			}
		} // OpenWrite

		public static bool OpenAppend(ref StreamWriter outF, string fName)	// open file for Writing
		{
			try
			{
				// open fName for Append as outF
				outF = new StreamWriter(fName, true, Encoding.ASCII);		//Appending = true
				return true;
			}
			catch (Exception ex)
			{
				MultiID.WriteDLn("Cannot open file \"" + fName + "\" for output");
				WCurDir(fName);
				ReportError("OpenAppend", ex.Message);
				return false;
			}
		} // OpenAppend

		public static void ReportError(string who, string errorMessage) // report IOResult etc. error, errorMessage = caught.Message
		{
			GC.Collect();	// force garbage collection of all generations
			MultiID.muErr = true;
			if (errorMessage == "") return;
			if (MultiID.muErrRep) MultiID.WriteTLn(who + ": " + errorMessage);
		}

		public static void ChangeDir(ref string st)
		// change the default disk directory and remove its name from command line
		{
			string stL;
			string savDir;
			if (st.Length != 0 && st[0] != '"')
				st = "\"" + st + "\"";
			if (st.Length == 0) st = "\""; // for safety
			stL = FileNm(ref st);
			savDir = "";
			try
			{
				if (!quiet) DoubleSpace();
				savDir = Environment.CurrentDirectory.ToString();
				if (stL == savDir) stL = "";
				if (stL != "") Directory.SetCurrentDirectory(stL);
			}
			catch
			{
				// -------- Trap for error handling
				MultiID.WriteLn("Path not found!");
			}
			finally
			{
				if (!quiet)
				{
					if (Environment.CurrentDirectory.ToString() != savDir)
					{
						MultiID.WriteTLn("Directory changed to \"" + Environment.CurrentDirectory.ToString() + '"');
					}
					else
					{
						MultiID.WriteTLn("Directory not changed \"" + Environment.CurrentDirectory.ToString() + '"');
					}
					MultiID.WriteTLn("Home directory is \"" + homeDir + '"');
				}
			}
		} // ChangeDir

		public static string FileNm(ref string st)
		// return and remove file name from command line
		{
			int i;
			string stL;
			if (st.Length == 0)
			{
				return "";
			}
			i = st.IndexOf("\"", 1);
			if (i <= 0) i = st.Length;
			stL = st.Substring(1, i - 1);
			DelNPBls(ref st, i); // do not delete the trailing " if any
			return stL;
		}

		public static void FixFName(ref string fName, string ext)
		// fix file name to path + name + .extension
		{
			fName = fName.Trim(' ');					// remove leading and trailing blanks
			int i = fName.IndexOf(" ");
			if (i >= 0) fName = fName.Substring(0, i);	// get rid of all blanks
			fName = fName.TrimEnd('.');					// remove trailing '.'
			if (fName == "") return;
			int p = fName.IndexOf(@"\");				// find last "\"
			i = fName.IndexOf('.', p + 1);				// a "." after the last "\" starts extension
			if (i < 0 && ext != "")
				fName = fName + "." + ext;				// use default extension
			i = fName.IndexOf(":.");					// :. OR \. is a bad file name
			if (i < 0) i = fName.IndexOf(@"\.");
			if (i >= 0) fName = fName.Substring(0, i);	// if \.xxx, or :.xxx, .xxx deleted
		}

		public static void WCurDir(string fName)
		// write current disk directory with final \
		// using Environment.CurrentDirectory.ToString(): C:\ already has the \ but non-root directories do not
		{
			// this can garble some output, but good info
			string cD = Directory.GetCurrentDirectory();
			if (cD.Substring(cD.Length - 1) != @"\") cD += @"\";
			if (fName.Length > 1 && fName[1] == ':')
			{
				MultiID.WriteTLn("Directory name = \"" + cD + '"');
			}
			else
			{
				MultiID.WriteTLn("Full name = \"" + cD + fName + '"');
			}
		}

		public static void DelNPBls(ref string st, int n)
		// delete 1st n characters plus new leading blanks, Side effect: noChange = false;
		{
			noChange = false;
			if (n >= st.Length) { st = ""; return; }
			st = st.Substring(n).TrimStart(' ');
		}


		public static void DisplayPriority() // display priority status
		{
			string st = Thread.CurrentThread.Priority.ToString();
			MultiID.WriteSLn("Priority is " + st);
		}

		public static void SetPriority(int p) // change priority to p
		{
			if (p == 2)
				Thread.CurrentThread.Priority = ThreadPriority.Highest;
			else if (p == 1)
				Thread.CurrentThread.Priority = ThreadPriority.AboveNormal;
			else if (p == -1)
				Thread.CurrentThread.Priority = ThreadPriority.BelowNormal;
			else if (p == -2)
				Thread.CurrentThread.Priority = ThreadPriority.Lowest;
			else
			{
				p = 0;
				Thread.CurrentThread.Priority = ThreadPriority.Normal;
			}
			pri = p;
		}

		public static void ReDimPreserve(ref int[] v, int n)
		// re-dimension array but preserve old values, n is new length
		// v must be ref because its pointer is changed by v = new int[n]
		{
			if (n == v.GetLength(0)) return;
			int m = Math.Min(n, v.GetLength(0));	// copy m values, m <= n
			try
			{
				int[] tV = new int[m];				// if n is smaller than old length,
				for (int j = 0; j < m; j++)			// upper values are lost
				{
					tV[j] = v[j];					// save m values in a temp array
				}
				v = new int[n];						// re-dimention
				for (int j = 0; j < m; j++)
				{
					v[j] = tV[j];					// restore m values
				}
			}
			catch (Exception ex)
			{
				long totalMemory = GC.GetTotalMemory(false);
				string st = ex.Message + ", M = " + n + ", v set to zero, Memory used = " + totalMemory;
				v = new int[2];
			}
		} // ReDimPreserve

		public static void ShiftL(long iL, MultiSI z) // z = z, iL "digits" more
		// Sets muErr true if overflow
		{
			int i;
			if (iL == 0) return;
			if (iL < 0)
			{
				ShiftR(-iL, z);
				return;
			}
			if (iL > Int32.MaxValue) iL = Int32.MaxValue;
			if (iL < Int32.MinValue) iL = Int32.MinValue;
			int iLi = (int)iL;
			if (iLi > (z.M - z.N))
			{
				MultiID.MuWriteErr("Signed Integer shift left overflow, continuing...");
				MultiID.ClearSI(z);
				return;
			}
			for (i = (z.N - 1); i >= 0; i--)
			{
				z.V[i + iLi] = z.V[i];
			}
			for (i = 0; i < iLi; i++)
			{
				z.V[i] = 0;
			}
			z.N += iLi;
		} // ShiftL

		public static void ShiftR(long iR, MultiSI z) // z = z, iR "digits" less
		// Sets muErr true if overflow
		{
			int i;
			if (iR == 0) return;
			if (iR < 0)
			{
				ShiftL(-iR, z);
				return;
			}
			if (iR > Int32.MaxValue) iR = Int32.MaxValue;
			if (iR < Int32.MinValue) iR = Int32.MinValue;
			int iRi = (int)iR;
			if (iRi >= z.N)
			{
				MultiID.ClearSI(z);
				return;
			}
			for (i = 0; i < z.N - iRi; i++)
			{
				z.V[i] = z.V[i + iRi];
			}
			for (i = z.N - iRi; i < z.N; i++)
			{
				z.V[i] = 0;
			}
			z.N -= iRi;
		} // ShiftR

		//public static void WriteSIF(MultiSI x, string name) // for testing
		//{
		//    StreamWriter outF = null;
		//    if (Common.OpenWrite(out outF, name + ".XIN"))
		//    {
		//        int tot = 0;
		//        int width = 65;
		//        bool WriteBase = true;
		//        MultiID.WriteTLn(0, outF, nL + name + " =" + nL, ref tot);
		//        tot = 0;
		//        MultiID.WriteSILn(0, outF, x, ref tot, width, WriteBase);
		//        outF.Close();
		//    }
		//}

		public static bool noChange;	// true after GetField until DelNPBls is called
		public static bool quiet;		// true if "quiet" mode is on
		public static int fMC;			// current max # of super digits in a result
		public static int nMax;			// max N for numbers on the list, fixed at startup
		public static int pri;			// execution priority [-2, 2]
		public static string homeDir;	// initial disk directory when program started
		public static string nL;		// one Environment.NewLine \n or \r or \r\n
	}
}
