// Juggler.cs : C# driver for JugglerC

//			JugglerW - Generate data for 50 dollar award

using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Text;			// for Encoding I/O to a Text file
using System.Reflection;	// for Assembly, ...
using System.Threading;		// for Thread

namespace MyNamespace
{
	/// <summary>
	///			JugglerC - Generate data for 50 dollar award, Console version
	///			Developed in Borland C++ 2.0.  Converted to Borland C++ 3.0 for Windows.
	///			Recompiled with Borland C++ 3.1 for Windows.
	///			and now in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>
	
	class Juggler
	{
		// The main entry point for the application JugglerC
		[STAThread]
		static int Main(string[] args)
		{
			Common.SetPriority(-1); // lower priority to BelowNormal
			try
			{
				Console.SetBufferSize(80, 1000);	// allow scrolling of output window
			}
			catch { }
			GetAssemblyAttributes();
			Common.homeDir = new DirectoryInfo(Environment.CurrentDirectory).ToString();
			logg = false;
			outT = false;
			Common.quiet = true;
			quit = false;
			restore = false;
			nL = Environment.NewLine;
			arg0 = "";
			if (args.Length > 0)				// if arguments
				arg0 = args[0];					// save 1st argument
			Init();
			MultiID.allowFHT = true;
			MultiID.debug = false;
			MultiID.diagOn = false;
			MultiID.muErrRep = true;
			MultiID.forceFHT = false;
			MultiID.timeIt = true;
			MultiID.writeBool = false;
			if (arg0 != "")						// if arguments
			{
				fileX = arg0;					// 1st arg is name of FileX
				arg0 = "";
				stIn = "Ex";
				WriteLn(nL + "1St Command line argument = " + fileX);
				GetCommand();
				DoCommand();
			}
			stIn = "";
			RunJuggler();
			if (logg)
				ToggleLogg();
			return 0;
		} // Main

		// ------- initialize/restart
		private static void Init()
		{
			MultiID.muNMax = 16777300;  // max number of super digits in a multi-precision
			// number. Allows for 134,218,400 = 2^27 + 672 decimal digits
			// each super digit is from 0 to 99,999,999, i.e. base = 10^8
			Common.nMax = MultiID.muNMax;
			Common.fMC = MultiID.muNMax;
			MultiID.intCalc = true;
			MultiID.outS = 1;
			new MultiID();				// initialize class MultiID
			new Common();				// initialize class Common
			new FHTMult();				// initialize class FHTMult
			new XIMult();				// initialize class XIMult
			x0 = new MultiSI();
			maxX = new MultiSI();
			maxX0 = new MultiSI();
			maxS0 = new MultiSI();
			MultiID.ValueSI10("48443", x0);
			maxI = 0;
			maxX &= 0;
			maxX0 &= 0;
			maxOut = true;
			fileRNew = true;
			maxS = -1;
			maxS0 &= 0;
			maxSOut = true;
			fileSNew = true;
			Directory.SetCurrentDirectory(Common.homeDir);
			string author1 = "Copyright (c) 1981-2008 by author: Harry J. Smith,";
			string author2 = "19628 Via Monte Dr., Saratoga, CA 95070.  All rights reserved.";
			string usage1 = "Usage: JugglerC [command-file-name-to-execute]";
			Console.Clear();
			WriteLn();
			WriteLn(title);		// show credits
			WriteLn(version);
			WriteLn(author1);
			WriteLn(author2);
			WriteLn();
			WriteLn(usage1);		// show usage
			WriteLn();
			Common.DisplayPriority();
			WriteLn();
			fileL = "FileL.Log";	// name of FileL
			fileM = "FileM.Max";	// name of FileM
			fileO = "FileO.Out";	// name of FileO
			fileR = "FileR.Sum";	// name of FileR
			fileS = "FileS.Sum";	// name of FileS
			fileX = "FileX.Code";	// name of FileX
			if (arg0 == "")
			{
				WriteLn("Help");
				Help();
				WriteLn();
			}
			if (arg0 == "")
			{
				WriteLn(nL + "State" + nL);
				State();
			}
		} // init

		public static void RunJuggler()	// get instructions from operator
		{
			while (!quit)
			{
				WriteLn();
				stIn = Console.ReadLine();
				WriteLog(stIn + nL);
				GetCommand();
				DoCommand();
				MultiID.muAbort = false;
				MultiID.muErr = false;
			}
		} // RunJuggler

		public static void DoCommand()	// do a command
		{
			if (stIn == "")	// "" = NoOp
				return;
			if (name == "AUTO+" && param == "")
				AutoPlus1();
			else if (name == "AUTOO" && param == "")
				AutoOdd();
			else if ((name == "CHDIR" || name == "CD") && param != "")
				Common.ChangeDir(ref param);
			else if ((name == "CLEARLOG" || name == "CL") && param == "")
				ClearLog();
			else if (name == "DIAG" && param != "")
				Diag();
			else if (name == "DF" && param == "")
				DiagF();
			else if (name == "DT" && param == "")
				DiagT();
			else if ((name == "EXECUTE" || name == "EX") && param == "")
				Execute();
			else if ((name == "EXIT" || name == "Q") && param == "")
				Exit();
			else if ((name == "FILEL" || name == "FL") && param != "")
				FileL();
			else if ((name == "FILEM" || name == "FM") && param != "")
				FileM();
			else if ((name == "FILEO" || name == "FO") && param != "")
				FileO();
			else if ((name == "FILEX" || name == "FX") && param != "")
				FileX();
			else if ((name == "HELP" || name == "HE") && param == "")
				Help();
			else if ((name == "INIT" || name == "IN") && param == "")
				Init();
			else if (name == "LF" && param == "")
				LogF();
			else if (name == "LOG" && param != "")
				Log();
			else if (name == "LT" && param == "")
				LogT();
			else if (name == "OF" && param == "")
				OutF();
			else if (name == "OT" && param == "")
				OutT();
			else if (name == "OUT" && param != "")
				Out();
			else if (name == "QF" && param == "")
				QuietF();
			else if (name == "QT" && param == "")
				QuietT();
			else if (name == "QUIET" && param != "")
				Quiet();
			else if ((name == "RESTORE" || name == "RE") && param == "")
				Restore();
			else if (name == "RUN" && param == "")
				RunSequence();
			else if (name == "RUNO" && param == "")
				RunOdd();
			else if (name == "RUN+" && param == "")
				RunPlus1();
			else if ((name == "SAVE" || name == "SA") && param == "")
				Save();
			else if ((name == "STATE" || name == "ST") && param == "")
			{
				WriteLn();
				State();
			}
			else if (name == "X0" && param != "")
				X0();
			else if ((name == "WRITEMAX" || name == "WM") && param == "")
				WriteMax();
			else if (stIn.Length > 1 && stIn.Substring(0, 2) == "//")
				return;
			else if (stIn.Length > 0 && stIn.Substring(0, 1) == "(")
				return;
			else
				WriteLn("I do not understand '" + stIn + "'");
		} // DoCommand

		public static void GetCommand()	// get name and parameter in command string
		{
			stIn = stIn.Trim();	// trim all white space from start and end of command
			string st = stIn;
			int at = st.IndexOf("//");
			if (at >= 0)
			{
				st = st.Substring(0, at);
				st = st.Trim();			// trim again
			}
			at = st.IndexOf("=");
			if (at >= 0)
			{
				name = st.Substring(0, at);
				name = name.Trim();
				name = name.ToUpper();
				if (name == "KEY")
					st = stIn;
				int len = st.Length;
				param = st.Substring(at + 1, len - at - 1);
				param = param.Trim(' ');
				if (name != "KEY")
				{
					param = param.Trim(' ', '"');
					at = param.IndexOf('"');
					if (at >= 0)
					{
						param = param.Substring(0, at);
						param = param.Trim();			// trim again
					}
					if (param == "")
						name = "NoName";
				}
			}
			else
			{
				name = st.ToUpper();
				param = "";
			}
		} // GetCommand

		public static void AutoOdd()	// Repeatedly do RunO to generate max x
		{
			RunSequence();
			if (MultiID.muAbort || MultiID.muErr)
				return;
			if (x0 == 1)
			{
				RunPlus1();				// do x0 = 2 which causes a max x and a max s
				if (MultiID.muAbort || MultiID.muErr)
					return;
			}
			while (true)
			{
				RunOdd();
				if (MultiID.muAbort || MultiID.muErr)
					return;
			}
		} // AutoOdd

		public static void AutoPlus1()	// Repeatedly do Run+ to generate max s
		{
			RunSequence();
			if (MultiID.muAbort || MultiID.muErr)
				return;
			while (true)
			{
				RunPlus1();
				if (MultiID.muAbort || MultiID.muErr)
					return;
			}
		} // AutoPlus1

		public static void ClearLog()	// clear Log file
		{
			DateTime dt = DateTime.Now;
			string dateTimeString = dt.ToString("MM/dd/yyyy hh:mm:ss tt");
			bool logS = logg;	// save Log status in a local
			if (logg)
			{
				string st = "Log file \"" + fileL + "\" Closed " + dateTimeString;
				if (!Common.quiet)
				{
					WriteLn(st);
					outL.WriteLine();
				}
				outL.WriteLine("");
				outL.Close();	// close the file
				logg = false;
			}
			// open log file for output as outL, clear the file, outL gotten by open routine
			if (Common.OpenWrite(out outL, fileL))
			{
				logg = true;	// set to get message to the file
				string st = "Log file \"" + fileL + "\" Cleared " + dateTimeString;
				if (!Common.quiet)
				{
					WriteLn(st);
					Common.WCurDir(fileL);
				}
				else
				{
					outL.WriteLine(st);
				}
				outL.WriteLine(title);
				outL.WriteLine(version);
				if (system != "") outL.WriteLine("Run on: " + system);
				if (!logS)
				{
					outL.Close();	// close the file
					logg = false;
				}
			}
		} // ClearLog

		public static void Execute()	// Execute command file FileX
		{
			stIn = "";
			StreamReader inF = null;
			if (!Common.OpenRead(out inF, fileX, out dir))
				return;
			stIn = inF.ReadLine();		// get first line to execute
			while (stIn != null)		// loop until end of file
			{
				WriteLn();
				if (stIn != "")
				{
					WriteLn(stIn);
					GetCommand();
					DoCommand();
				}
				stIn = inF.ReadLine();	// get the next line to execute
			}
			inF.Close();
		} // Execute

		public static void Exit()	// Exit/Quit the program
		{
			quit = true;
		} // Exit

		public static void FileL()	// Set file name of Log file
		{
			Common.FixFName(ref param, "Log");
			if (param == "")
				WriteLn("Illegal file name");
			else
				fileL = param;	// name of FileL
		} // FileL

		public static void FileM()	// Set file name of MaxX file
		{
			Common.FixFName(ref param, "Max");
			if (param == "")
				WriteLn("Illegal file name");
			else
				fileM = param;	// name of FileM
	} // FileM

		public static void FileO()	// Set file name of Output file
		{
			Common.FixFName(ref param, "Out");
			if (param == "")
				WriteLn("Illegal file name");
			else
				fileO = param;	// name of FileO
		} // FileO

		public static void FileX()	// Set file name of eXecute file
		{
			Common.FixFName(ref param, "Code");
			if (param == "")
				WriteLn("Illegal file name");
			else
			{
				fileX = param;	// name of FileX
				if (!restore)
					VerifyF();
			}
		} // FileX

		public static void VerifyF()	// verify that File param exists
		{
			StreamReader inF = null;
			if (Common.OpenRead(out inF, param, out dir))    // Try to open new File param
			{
				inF.Close();
				if (!Common.quiet)
					WriteLn("File \"" + param + "\" verified and closed");
			}
		} // VerifyF

		public static void Help()	// Show Help file JugglerCHelp.txt
		{
			string st = "";
			StreamReader inF = null;
			if (!Common.OpenRead(out inF, "JugglerCHelp.txt", out dir))
				return;
			WriteLn();
			st = inF.ReadLine();	// get the next line of Help
			while (st != null)	// loop until end of file
			{
				WriteLn(st);
				st = inF.ReadLine();	// get the next line of Help
			}
			inF.Close();
		} // Help

		public static void Log()	// Set Log screen to log file mode on or off
		{
			bool loggL = param != "0";
			if (logg != loggL)
				ToggleLogg();
		} // Log

		public static void LogT()	// Set Log screen to log file mode on (True)
		{
			if (!logg)
				ToggleLogg();
		} // LogT

		public static void LogF()	// Set Log screen to log file mode off (False)
		{
			if (logg)
				ToggleLogg();
		} // LogF

		public static void Out()	// Set Log screen to log file mode on or off
		{
			outT = param != "0";
		} // Out

		public static void OutT()	// Set Log screen to log file mode on (True)
		{
			outT = true;
		} // OutT

		public static void OutF()	// Set Log screen to log file mode off (False)
		{
			outT = false;
		} // OutF

		public static void Diag()	// Set Diagnostic mode on or off
		{
			MultiID.diagOn = param != "0";
		} // Diag

		public static void DiagT()	// Set Diagnostic mode on (True)
		{
			MultiID.diagOn = true;
		} // DiagT

		public static void DiagF()	// Set Diagnostic mode off (False)
		{
			MultiID.diagOn = false;
		} // DiagF

		public static void Quiet()	// Set quiet mode on or off
		{
			Common.quiet = param != "0";
		} // Quiet

		public static void QuietT()	// Set quiet mode on (True)
		{
			Common.quiet = true;
		} // QuietT

		public static void QuietF()	// Set quiet mode off (False)
		{
			Common.quiet = false;
		} // quietF

		public static void Restore()	// Restore the state of inputs from Save.Code
		{
			StreamReader inF = null;
			if (!Common.OpenRead(out inF, "Save.Code", out dir))
				return;
			restore = true;
			stIn = inF.ReadLine();	// get the next line of Save.Code
			while (stIn != null)	// loop until end of file
			{
				GetCommand();
				DoCommand();
				stIn = inF.ReadLine();	// get the next line of Save.Code
			}
			inF.Close();
			WriteLn();
			State();
			restore = false;
		} // Restore

		public static void Save()	// Save the state of inputs to Save.Code
		{
			string stD = "0";
			if (MultiID.diagOn)
				stD = "1";
			string stL = "0";
			if (logg)
				stL = "1";
			string stO = "0";
			if (outT)
				stO = "1";
			string stQ = "0";
			if (Common.quiet)
				stQ = "1";
			StreamWriter outF = null;
			if (!Common.OpenWrite(out outF, "Save.Code"))
				return;
			if (!WriteFSt(outF, "ChDir = " + Environment.CurrentDirectory + nL))
			{
				outF.Close();
				return;
			}
			WriteFSt(outF, "Diag = " + stD + nL);
			WriteFSt(outF, "FileL = " + QuoteFN(fileL) + nL);
			WriteFSt(outF, "FileM = " + QuoteFN(fileM) + nL);
			WriteFSt(outF, "FileO = " + QuoteFN(fileO) + nL);
			WriteFSt(outF, "FileX = " + QuoteFN(fileX) + nL);
			WriteFSt(outF, "Log = " + stL + nL);
			WriteFSt(outF, "Out = " + stO + nL);
			WriteFSt(outF, "Quiet = " + stQ + nL);
			int tot = 0;
			int width = Int32.MaxValue;
			bool WriteBase = true;
			MultiID.WriteT(0, outF, "x0 = ", ref tot);
			MultiID.WriteSILn(0, outF, x0, ref tot, width, WriteBase);
			outF.Close();
			if (!Common.quiet)
				WriteLn("Save file \"Save.Code\" updated");
		} // Save

		public static string QuoteFN(string fileName)	// quote fileName if needed
		{
			string delim = ";=/:()@\"";	// to be consistent
			if (fileName.Contains(" "))
				return '"' + fileName + '"';
			for (int i = 0; i < delim.Length; i++)
			{
				if (fileName.Contains(delim[i].ToString()))
					return '"' + fileName + '"';
			}
			return fileName;
		} // QuoteFN

		public static void State()	// Show state of inputs
		{
			string stD = "0";
			if (MultiID.diagOn)
				stD = "1";
			string stL = "0";
			if (logg)
				stL = "1";
			string stO = "0";
			if (outT)
				stO = "1";
			string stQ = "0";
			if (Common.quiet)
				stQ = "1";
			WriteLn(" Home Dir: \"" + Common.homeDir + '"');
			WriteLn("File path: \"" + Environment.CurrentDirectory.ToString() + '"');
			WriteLn();
			WriteLn("  Diag = " + stD);
			WriteLn(" FileL = " + QuoteFN(fileL));
			WriteLn(" FileM = " + QuoteFN(fileM));
			WriteLn(" FileO = " + QuoteFN(fileO));
			WriteLn(" FileX = " + QuoteFN(fileX));
			WriteLn("   Log = " + stL);
			WriteLn("   Out = " + stO);
			WriteLn(" Quiet = " + stQ);
			int tot = 0;
			int width = Int32.MaxValue;
			bool WriteBase = true;
			MultiID.WriteT(1, null, "    x0 = ", ref tot);
			MultiID.WriteSILn(x0, ref tot, width, WriteBase);	// write xo to Console
		} // State

		public static void X0()	// Set starting number x(0) to x
		{
			int used;
			ValueN(param, out used, x0);
			if (x0 < MultiID.sOne)
				x0 &= 1;
		} // X0

		public static void ValueN(string st, out int i, MultiSI z)
		// convert String to z, returns i = number of character used
		// allows st to be a literal, muErr set true if overflow
		{
			int m;
			m = (int)(5 + st.Length * Math.Log10(MultiID.baseI) / MultiID.muDMax);
			if (m > Common.nMax) m = Common.nMax;
			MultiSI t = new MultiSI(m);
			MultiID.ValueSI(st, out i, t);
			z &= t; 
		}

		public static void ToggleLogg() //Toggle Logging of output screen to Log File
		{
			DateTime dt = DateTime.Now;
			string dateTimeString = dt.ToString("MM/dd/yyyy hh:mm:ss tt");
			if (logg)
			{
				string st = "Log file \"" + fileL + "\" Closed " + dateTimeString;
				if (!Common.quiet)
				{
					WriteLn(st);
					outL.WriteLine();
				}
				outL.Close();	// close the file
				logg = false;
			}
			else
			{
				if (Common.OpenAppend(ref outL, fileL)) // open fileL for Append as outL
				{
					logg = true;
					string st = "Log file \"" + fileL + "\" Opened for Append " + dateTimeString;
					if (!Common.quiet)
					{
						WriteLn(st);
						Common.WCurDir(fileL);
						outL.WriteLine(title);
						outL.WriteLine(version);
						if (system != "") outL.WriteLine("Run on: " + system);
					}
				}
				else
				{
					return;
				}
			}
		} // ToggleLogg

		public static void WriteLn() // Write linefeed
		{
			Write(nL);
		}

		public static void WriteLn(string st) // Write st plus linefeed
		{
			Write(st + nL);
		}

		public static void Write(string st) // print string to console and Log file
		{
			Console.Write(st);
			WriteLog(st);
		} // Write

		public static void WriteLog(string st) // print string to Log file
		{
			if (logg)
			{
				try { outL.Write(st); }
				catch { }
			}
		} // WriteLog

		public static bool WriteFSt(StreamWriter outF, string st)	// write string to an opened file
		{
			if (st == "")
				return false;
			if (outF == null)
				return false;
			try
			{
				outF.Write(st);
				return true;
			}
			catch (Exception ex)
			{
				WriteLn("File write error " + ex.Message + nL);
				return false;
			}
		} // WriteFSt

		public static void RunOdd()	// Increase x(0) to next odd number and generate the Sequence
		{
			if ((x0.V[0] & 1) == 0) // if even
				x0 += 1;
			else
				x0 += 2;
			RunSequence();
		}

		public static void RunPlus1()	// Increase x(0) to next odd number and generate the Sequence
		{
			x0 += 1;
			RunSequence();
		}

		public static void TestAbort()	// abort by operator?
		{
			if (Console.KeyAvailable)
			{
				string st;
				WriteLn("Calculations stopped");
				WriteLn("To abort calculations, Press A <Enter>, else press <Enter>");
				char ch = ' ';
				st = Console.ReadLine();
				if (st != "") ch = st[st.Length - 1];
				if (ch == 'A' || ch == 'a')
					MultiID.muAbort = true;
				else
					WriteLn("Continuing calculations");
			}
		}

		public static void ReSizeN(int min, MultiSI x, MultiSI z) // size z = size x
		// z may be left un-normalized
		{
			int zM;
			int xM = x.M;
			if (xM < min) xM = min;
			zM = z.M;
			if (xM == zM) return;
			Common.ReDimPreserve(ref z.V, xM + 1);
			z.M = xM;
			if (z.N > xM) z.N = xM;
		}

		//public static bool TestGoodSI(MultiSI x)	// for testing
		//{
		//    for (int i = 0; i < x.N; i++)
		//        if (x.V[i] < 0)
		//            return false;
		//    return true;
		//}

		public static void RunSequence()
		{
			string st;
			int tot;
			int L = 0;		// Length of sequence
			int LSD;		// Least significant super digit
			MultiSI x;		// Current x
			MultiSI x2;		// Current x squared
			x = new MultiSI(4);
			x2 = new MultiSI(4);
			MultiID.tV0 = MultiID.TimerF();
			DateTime dt = DateTime.Now;
			string dateTimeString = dt.ToString("MM/dd/yyyy hh:mm:ss tt");
			if ((x0 == MultiID.sOne || (x0.V[0] & 1) == 0) && x0 > maxX)	// if x0 is even and > max found so far
			{
				maxTime = dateTimeString;
				maxX &= x0;
				maxI = 0;
				maxX0 &= x0;
				maxOut = false;
			}
			if (outT)
			{
				// open output file for output as outO, clear the file, outO gotten by open routine
				if (Common.OpenWrite(out outO, fileO))
				{
					st = "Output file \"" + fileO + "\" Cleared " + dateTimeString;
					if (!Common.quiet)
					{
						WriteLn(st);
						Common.WCurDir(fileO);
					}
				}
				else
				{
					WriteLn("Cannot open " + fileO);
					return;
				}
			}
			WriteLn();
			st = "Juggler Sequence starting at " + dateTimeString;
			if (outT)
			{
				outO.WriteLine(title);
				outO.WriteLine(version);
				if (system != "") outO.WriteLine("Run on: " + system);
				outO.WriteLine();
				outO.WriteLine(st);
			}
			WriteLn(st);
			st = "with x(0) = ";
			tot = st.Length;
			int width = Int32.MaxValue;
			bool WriteBase = true;
			if (outT)
			{
				outO.Write(st);
				MultiID.WriteSI(0, outO, x0, ref tot, width, WriteBase);
				outO.WriteLine();
				outO.WriteLine();
			}
			Write(st);
			MultiID.WriteSI(1, null, x0, ref tot, width, WriteBase);
			WriteLn();
			WriteLn();
			st = "x(i) = x (digits in x)" + nL;
			if (outT)
			{
				outO.WriteLine(st);
				st = "x(0) = ";
				outO.Write(st);
				tot = st.Length;
				MultiID.WriteSILn(0, outO, x0, ref tot, width, WriteBase);
			}
			WriteLn("N = number of digits in x, T = running time, DT = delta time");
			WriteLn("To abort calculations, Press A <Enter> or Ctrl-C");
			WriteLn();
			x &= x0;
			shortD = 42;	// Select short format for output, at most 42 in long format
			MultiID.tV2 = MultiID.tV1 = MultiID.TimerF();	// Reset timing from zero
			while (((x0 == MultiID.sOne) || (x > MultiID.sOne)) && !MultiID.muErr && !MultiID.muAbort && !MultiID.softAbort)
			{
				//if(!TestGoodSI(x))	// for testing
				//{
				//    Common.ReportError("RunSequence", "Negative number in a super digit");
				//    break;
				//}
				LSD = x.V[0];
				if (LSD % 2 == 1)
				{
					Write("U  Up started  ");
					if (MultiID.diagOn)
					{
						Write("  ");
						MultiID.Diag("");
						WriteLn();
					}
					x2 &= x * x;
					TestAbort();
					//if (!TestGoodSI(x2))	// for testing
					//    Common.ReportError("RunSequence", "Negative number in a super digit");
					if (MultiID.muAbort || MultiID.muErr) break;
					if (MultiID.diagOn)
					{
						Write("  x^2 ended  ");
						Write("    ");
						MultiID.Diag("");
						WriteLn();
					}
					x &= x * x2;
					TestAbort();
					//if (!TestGoodSI(x))	// for testing
					//    Common.ReportError("RunSequence", "Negative number in a super digit");
					if (MultiID.muAbort || MultiID.muErr) break;
					if (MultiID.diagOn)
					{
						Write("  x * x^2 ended  ");
						MultiID.Diag("");
						WriteLn();
					}
					MultiID.RSqRtSI(x);
				}
				else {
					Write("D  Down started");
					if (MultiID.diagOn)
					{
						Write("  ");
						MultiID.Diag("");
						WriteLn();
					}
					MultiID.RSqRtSI(x);
					//MultiID.SqRootSI(x, x2);	// for testing
					//if (!TestGoodSI(x2))
					//{
					//    Common.WriteSIF(x, "BadX2");
					//    Common.ReportError("RunSequence", "Negative number in a super digit");
					//}
					//x &= x2;
			}
				TestAbort();
				//if (!TestGoodSI(x))	// for testing
				//    Common.ReportError("RunSequence", "Negative number in a super digit");
				if (MultiID.muAbort || MultiID.muErr) break;
				if (MultiID.diagOn)
				{
					st = "  SqRt ended  ";
					Write(st);
					Write("   ");
					MultiID.Diag("");
					WriteLn();
				}
				L++;
				if (outT)
				{
					st = "x(" + L + ") = ";
					tot = st.Length;
					outO.Write(st);
					MultiID.ShortWrSILn(0, outO, shortD, x, ref tot);
				}
				Write("  i = " + L + "  N = " + DigitsIn(x) + "   ");
				MultiID.TimeIt("");
				WriteLn();
				if ((x.V[0] & 1) == 0 && x > maxX)	// if x is even and > max found so far
				{
					dt = DateTime.Now;
					maxTime = dt.ToString("MM/dd/yyyy hh:mm:ss tt");	// save max x found
					maxX &= x;
					maxI = L;
					maxX0 &= x0;
					maxOut = false;
				}
				if (outT)
				{
					outO.Close();						// close and reopen the output file, flush
					Common.OpenAppend(ref outO, fileO);
				}
				if (logg)
				{
					outL.Close();					// close and reopen the log file, flush
					Common.OpenAppend(ref outL, fileL);
				}
				if (x0 == MultiID.sOne)
				{
					L--;
					break;
				}
			}
			if (MultiID.muErr) {
				st = "Run aborted due to error in calculations";
				if (outT) outO.WriteLine(st);
				WriteLn(st);
			}
			else if (MultiID.muAbort) {
				st = "Run aborted by operator";
				if (outT) outO.WriteLine(st);
				WriteLn(st);
				maxI = 0;
				maxX &= 0;
				maxX0 &= 0;
				maxOut = true;
				fileRNew = true;
				maxS = -1;
				L = -1;
				maxS0 &= 0;
				maxSOut = true;
				fileSNew = true;
			}
			else if (MultiID.softAbort) {
				st = "Soft abort by operator";
				if (outT) outO.WriteLine(st);
				WriteLn(st);
			}
			else
			{
				st = "Length = " + (L+1) + ", Steps = " + L;
				if (outT) outO.WriteLine(st);
				WriteLn(st);
			}
			dt = DateTime.Now;
			dateTimeString = dt.ToString("MM/dd/yyyy hh:mm:ss tt");
			st = "Juggler Sequence ended at " + dateTimeString;
			if (outT)
			{
				outO.WriteLine();
				outO.WriteLine(st);
				outO.Close();	// close the Output file
			}
			WriteLn();
			WriteLn(st);
			if (!maxOut)
				WriteMax();
			if (L > maxS)		// if steps > maxS found so far
			{
				maxS = L;		// save max steps found
				maxS0 &= x0;
				maxSOut = false;
			}
			if (!maxSOut)
				WriteMaxS();
			if (logg)
			{
				outL.Close();					// close and reopen the log file
				Common.OpenAppend(ref outL, fileL);
			}
		} // RunSequence

		public static void WriteMax()	// Output Max Juggler Number computed to MaxX file
		{
			StreamWriter outM = null;
			if (Common.OpenWrite(out outM, fileM))
			{
				string st;
				int tot;
				int width;
				bool WriteBase;
				st = "x(0) = ";
				outM.Write(st);
				tot = st.Length;
				width = Int32.MaxValue;
				WriteBase = true;
				MultiID.WriteSI(0, outM, maxX0, ref tot, width, WriteBase);
				outM.WriteLine();
				outM.WriteLine();
				outM.WriteLine("x(" + maxI + ") =");
				tot = 0;
				width = 65;
				WriteBase = true;
				MultiID.WriteSILn(0, outM, maxX, ref tot, width, WriteBase);
				outM.WriteLine();
				outM.WriteLine("Computed by program:");
				outM.WriteLine(title);
				outM.WriteLine(version);
				if (system != "") outM.WriteLine("Run on: " + system);
				outM.WriteLine();
				st = "Juggler Sequence Max computed at " + maxTime;
				outM.WriteLine(st);
				st = "with x(0) = ";
				outM.Write(st);
				tot = st.Length;
				width = Int32.MaxValue;
				WriteBase = true;
				MultiID.WriteSI(0, outM, maxX0, ref tot, width, WriteBase);
				outM.WriteLine();
				outM.WriteLine();
				outM.Close();
				maxOut = true;
				WriteLn();
				StreamWriter outR = null;
				bool rOK = Common.OpenAppend(ref outR, fileR);
				if (fileRNew && rOK)
				{
					outR.WriteLine("-----------------------------------");
					fileRNew = false;
				}
				int digits = DigitsIn(maxX);
				st = "x(" + maxI + ") max with " + digits + " digits for x(0) = ";
				Write(st);
				if (rOK) outR.Write(st);
				tot = st.Length;
				MultiID.WriteSI(1, outR, maxX0, ref tot, width, WriteBase);
				WriteLn();
				if (rOK)
				{
					outR.WriteLine();
					outR.Close();
				}
			}
		} // WriteMax

		public static void WriteMaxS()	// Output Max Steps fileS file
		{
			string st;
			int tot;
			int width;
			bool WriteBase;
			width = Int32.MaxValue;
			WriteBase = true;
			StreamWriter outS = null;
			bool sOK = Common.OpenAppend(ref outS, fileS);
			if (fileSNew && sOK)
			{
				outS.WriteLine("-----------------------------------");
				fileSNew = false;
			}
			st = "Max steps = " + maxS + " for x0 = ";
			Write(st);
			if (sOK) outS.Write(st);
			tot = st.Length;
			MultiID.WriteSI(1, outS, maxS0, ref tot, width, WriteBase);
			WriteLn();
			if (sOK)
			{
				outS.WriteLine();
				outS.Close();
			}
			maxSOut = true;
		} // WriteMaxS

		public static int DigitsIn(MultiSI x)	// return number of decimal digits in x
		{
			int nn = x.N;
			string st = x.V[nn - 1].ToString();
			int digits = (nn - 1) * MultiID.muDMax + st.Length;
			return digits;
		} // DigitsIn

		public static void GetAssemblyAttributes()	// and version
		{
			system = Environment.GetEnvironmentVariable("SYSTEM");
			if (system == null)
				system = "";
			else
				system += Environment.NewLine;
			system += Environment.GetEnvironmentVariable("COMPUTERNAME");
			system += " by ";
			system += Environment.GetEnvironmentVariable("USERNAME");
			if (system == null || system == "")
				system = "<Computer system name not set>";
			title = "JugglerC - Generate data for 50 dollar award written in C#";
			Version aVersion = Assembly.GetExecutingAssembly().GetName().Version;
			version = "C# Console Version " + aVersion.ToString();
		} // GetAssemblyAttributes

		private static bool logg;		// true if Logging screen output to FILE
		private static bool outT;		// true if output sequence to Output file mode is on
		private static bool quit;		// Exit/Quit flag
		private static bool restore;	// Restore flag
		private static StreamWriter outL;	// log file output #
		private static StreamWriter outO;	// output file output #
		private static string arg0;	    // 1st command line argument
		private static string fileL;	// name of FileL, Log file
		private static string fileM;	// name of FileM, MaxX file
		private static string fileO;	// name of FileO, Output file
		private static string fileR;	// name of FileR, Summery of MaxX file
		private static string fileS;	// name of FileS, Summery of max Steps/Lengths file
		private static string fileX;	// name of FileX, eXecute file
		private static string name;		// name in command string
		private static string nL;		// one Environment.NewLine \n or \r or \r\n
		private static string param;	// parameter in command string
		private static string stIn;		// command string as input
		private static string system;	// EnvironmentVariable("SYSTEM")
		private static string title;	// program title
		private static string version;	// = "C# Console Version x.x.x.xxxxx"
		private static MultiSI x0;		// Starting x, x zero
		private static int shortD;		// If > 0 then Short format desired from x.Writ( Out)
		private static MultiSI maxX;	// max x found
		private static MultiSI maxX0;	// x0 of max x found
		private static int maxI;		// index of max x found
		private static string maxTime;	// date time string of max x found
		private static bool maxOut;		// true iff this maxX has been output
		private static bool fileRNew;	// true iff fileR has not been written to yet
		private static int maxS;		// max steps found
		private static MultiSI maxS0;	// x0 of max steps found
		private static bool maxSOut;	// true iff this maxS has been output
		private static bool fileSNew;	// true iff fileS has not been written to yet
		private static string dir;		// name of directory file opened in
	}
}
